import random
import sys
import numpy as np
from numpy.linalg import norm
from deap import base
from deap import creator
from deap import tools
from operator import attrgetter

######## major setting area ########
restrain_filename = 'ExpData'					## The input file of experimental data
initial_pool_file = [['matrix-A-', 2], ['matrix-B-', 3]]	## The output file list. Change accordingly.
## The example file names which going to be read in will be: matrix-A-0, matrix-A-1, matrix-B-0, matrix-B-1, matrix-B-2.
tPool_size = 320000				## the size of initial pool. The input number can be larger than actual size, but not smaller.						
eachPool_size = 1000			## Size of Part2 & Part3.
IterMax = 5						## Max iteration times when no new best ensemble appears
cxPB, mutPB, nGEN = 0.5, 0.2, 1000		## crossover rate, mutation rate, GA generation
tournament_sizepool = [50, 25, 20, 10, 5, 2, 1]		## Number of tournaments in one generation
							## It's value gradually decrease to increase slection pressure 
							## Can change the list length and its elements
							## Make sure the population size is divisible by all the elements
result = open('opfile','w')			## output file name.
enSize = 200				## ensemble size
POPU = 100					## population size used in GA

########################################################################
#                                                                      # 
# DO NOT change the following code, unless you know what you are doing!#
#                                                                      # 
########################################################################

toolbox = base.Toolbox()
######### read in initial pool data #######
def read_iniPool(nameList, tPool_size, prePoints):
	Index = np.zeros((tPool_size))
	sumMA = np.zeros((tPool_size, prePoints))
	count = 0
	for eachname in nameList:
		for blk in xrange(eachname[1]):
			name1 = '{}{}'.format(eachname[0], blk)
			print 'Loading:', name1
			matriA = open(name1,'r')
			for line in matriA.readlines():
				line = line.strip().split()
				if len(line) != prePoints + 1:
					print "Calculat Restraint number doesn't match with experimental data."
					sys.exit()
				Index[count] = int(float(line[0]))
				sumMA[count, :] = [float(i) for i in line[1:]]
				count += 1
			matriA.close()
	knorm = norm(sumMA, axis=1)
	zerosPos = np.where(knorm == 0)[0]
	Index = np.delete(Index, zerosPos, 0)
	sumMA = np.delete(sumMA, zerosPos, 0)
	print 'Shape of Initial Pool:', np.shape(sumMA)
	return Index, sumMA

######## read in experimental data ########
def read_exRST(file):
	matriB = open(file,'r')
	print 'Loading: experimental Restraint'
	exRST = []
	for line in matriB.readlines():
		line = line.strip().split()
		exRST.append(float(line[0]))
	matriB.close()
	print 'Restraint NO:', len(exRST)
	return exRST

######## fitness function ########
def evalPRE(individual, mpPool, sumMA, exRST):
	mpPoolindex = [mpPool[x] for x in individual]
	for i in mpPoolindex:
		if type(i) is not int:
			print i
 			sys.exit()
	indMA = sumMA[mpPoolindex, :]
	calRST = np.sum(indMA, axis=0) / enSize
	Q = norm(calRST - exRST) / norm(exRST)
	return Q,		# Always return a turple!
	
######## external muation operator ########
def ex_mutate(ind, low, up, indpb):
	flag = 1
	for i in xrange(len(ind)):
		if random.random() < indpb:
			ind[i] = random.randint(low, up)
			flag = 0
	if flag:
		site = random.randint(0, enSize-1)
		ind[site] = random.randint(low, up)
	return ind,
	
######## internal muation operator ########	
def in_mutate(ind, library, indpb):
	flag = 1
	for i in xrange(len(ind)):
		if random.random() < indpb:
			ind[i] = random.choice(library)
			flag = 0
	if flag:
		site = random.randint(0, enSize-1)
		ind[site] = random.choice(library)
	return ind,

######## genetic algorithm in DEAP ########
def gaInMPGA(mpPool, enSize):
	library = []
	offspring = toolbox.population(n=POPU)
	for g in xrange(nGEN):
		for ind in offspring:										# build library
			library += ind
		library = list(set(library))
		os_mate1 = list(map(toolbox.clone, offspring))				# mate
		for child1, child2 in zip(os_mate1[::2], os_mate1[1::2]):
			toolbox.mate(child1, child2)
		os_exmu = list(map(toolbox.clone, offspring))				# exmutation
		for mutant in os_exmu:
			toolbox.ex_mutate(mutant)
		os_inmu = list(map(toolbox.clone, offspring))				# inmutation	
		for mutant in os_inmu:
			toolbox.in_mutate(mutant, library=library)		
		mixture = offspring + os_mate1 + os_exmu + os_inmu	# combine
		fitnesses = list(map(toolbox.evaluate, mixture))
		for ind, fit in zip(mixture, fitnesses):
			ind.fitness.values = fit
		mixture.sort(key=attrgetter("fitness"))
		dup_count = 0												# duplication examination
		for i in xrange(len(mixture)-1):
			if mixture[i].fitness.values[0] != mixture[i+1].fitness.values[0]: continue
			toolbox.ex_mutate(mixture[i], indpb=0.1)
			del mixture[i].fitness.values
			dup_count += 1
		dups = [ind for ind in mixture if not ind.fitness.valid]
		fitnesses = map(toolbox.evaluate, dups)
		for ind, fit in zip(dups, fitnesses):
			ind.fitness.values = fit
		random.shuffle(mixture)
		offspring = []												# selection
		tour_num = tournament_sizepool[int(7 * g / nGEN)]
		tour_size = len(mixture) / tour_num
		for t in xrange(tour_num):
			submix = mixture[t * tour_size : (t + 1) * tour_size]
			submix.sort(key=attrgetter("fitness"), reverse=True)
			sel_submix = submix[:(POPU/tour_num)]
			offspring += sel_submix
		if g % 50 ==0:
			fits = [ind.fitness.values[0] for ind in offspring]
			mean = sum(fits) / POPU
			sum2 = sum(x*x for x in fits)
			std = abs(sum2 / POPU - mean**2)**0.5
			print ("Gen %i \tMin %.8s \tMax %.8s \tAvg %.8s \tStd %.8s"
					%(g, min(fits), max(fits), mean, std)
					)
			print 'tour_size', tour_size, 'tour_num', tour_num
			print 'Identical ensemble number in current generation:', dup_count
	bestID = np.argmin(fits)
	bestEn = offspring[bestID]
	return bestEn

def main():
	exRST = read_exRST(restrain_filename)
	prePoints = len(exRST)
	Index, sumMA = read_iniPool(initial_pool_file, tPool_size, prePoints)
##### norms #####
	norm2_expPRE = norm(exRST)
	norm1_expPRE = norm(exRST, ord=1)
	norm2_pool = norm(sumMA, axis=1)
	norm1_pool = norm(sumMA, ord=1, axis=1)
################ deap setting
	creator.create("FitnessMin", base.Fitness, weights=(-1.0,))
	creator.create("Individual", list, fitness=creator.FitnessMin)
	toolbox.register("mate", tools.cxUniform, indpb=cxPB)
	toolbox.register("in_mutate", in_mutate, indpb=mutPB)
################
	HOF = []	# hall of fame
	residual = exRST[:]
	iter = 0
	bestQ = 1
	while iter < IterMax:
		product = np.inner(sumMA, residual)    
		product1 = product / norm2_pool
		sel1 = np.argsort(-product1).tolist()[:eachPool_size]
		product2 = product1 - abs(norm1_pool - norm1_expPRE)
		sel2 = np.argsort(-product2).tolist()[:eachPool_size]
		mpPool = HOF + sel1 + sel2
		mpPoolsize = len(mpPool)
		toolbox.register("evaluate", evalPRE, mpPool=mpPool, sumMA=sumMA, exRST=exRST)
		toolbox.register("attr_no", random.randint, 0, mpPoolsize-1)
		toolbox.register("individual", tools.initRepeat, creator.Individual, toolbox.attr_no, enSize)
		toolbox.register("population", tools.initRepeat, list, toolbox.individual)
		toolbox.register("ex_mutate", ex_mutate, low=0, up=mpPoolsize-1, indpb=0.01)
			
		survival = gaInMPGA(mpPool, enSize)
		pos = [mpPool[x] for x in survival]
		HOF += pos
		calMatrix = sumMA[pos,:]
		calRST = calMatrix.sum(axis=0) / len(survival)
		tmp_residual = map(lambda (a,b): a -b, zip(exRST, calRST))
		Q = norm(tmp_residual) / norm2_expPRE
		if Q >= bestQ: 
			iter += 1
			continue
		bestQ = Q
		bestpos = pos
		residual = tmp_residual
		rlist =  'Iter: %d Q factor: %f Ensemble Size: %d Unique Conformers: %d' % (iter, Q, enSize, len(set(pos)))
		result.write(rlist + '\n')
	result.write('~~~~~~~~~~Postions of the optimal ensemble members:' + '\n')
	result.write(' '.join([str(i) for i in bestpos]) + '\n')
	pdbID = Index[bestpos]
	result.write('~~~~~~~~~~Struture ID of the optimal ensemble members:' + '\n')
	result.write(' '.join([str(int(i)) for i in pdbID]) + '\n')

if __name__ == '__main__':
	main()
